@@ -36,7 +36,7 @@ pub enum Error {
     EmptyCommitRange,
     #[error("There is no path from {0} to {1}")]
     NoPathConnecting(Oid, Oid),
-    #[error("Key `{0}` missing")]
+    #[error("Could not find a certificate with {0} in the policy file")]
     MissingKey(openpgp::KeyHandle),
     #[error("Key `{0}` is bad: {1}")]
     BadKey(openpgp::KeyHandle, String),
