@@ -31,6 +31,7 @@ mod macros;
 mod cli;
 use cli::CertArg;
 use cli::PolicyFileArg;
+use cli::Verbosity;
 use cli::paths::StateDirectory;
 
 mod commands;
@@ -111,6 +112,7 @@ pub struct Config<'a> {
     home: Option<sequoia_directories::Home>,
     cert_store_path: Option<StateDirectory>,
     cert_store: OnceCell<CertStore<'a>>,
+    verbosity: Verbosity,
 }
 
 impl<'a> Config<'a> {
@@ -200,6 +202,16 @@ impl<'a> Config<'a> {
         Ok(self.cert_store.get().expect("just configured"))
     }
 
+    /// Whether the verbosity is set to verbose.
+    pub fn verbose(&self) -> bool {
+        self.verbosity.verbose()
+    }
+
+    /// Whether the verbosity is set to quiet.
+    pub fn quiet(&self) -> bool {
+        self.verbosity.quiet()
+    }
+
     fn read_policy<'f, F>(&self, policy_file: F) -> Result<Policy>
     where
         F: Into<Option<&'f PolicyFileArg>>,
@@ -381,6 +393,7 @@ fn main() -> anyhow::Result<()> {
         home,
         cert_store_path: cli.cert_store,
         cert_store: Default::default(),
+        verbosity: Verbosity::new(cli.quiet, cli.verbose),
     };
 
     let commit_by_symbolic_name = |git: &git2::Repository, name: &str,
@@ -467,7 +480,7 @@ fn main() -> anyhow::Result<()> {
 
     match cli.subcommand {
         cli::Subcommand::Init(command) => {
-            commands::init::dispatch(command)?;
+            commands::init::dispatch(&config, command)?;
         }
 
         cli::Subcommand::Policy { command } => {
@@ -529,10 +542,11 @@ fn main() -> anyhow::Result<()> {
                                output::Commit::new(
                                    &git, oid, parent_oid,
                                    &policy_file.path, result)?
-                                   .describe(&mut io::stdout())?;
+                                   .describe(&mut io::stdout(), &config.verbosity)?;
                                Ok(())
                            },
                            &mut cache,
+                           config.quiet(), config.verbose(),
                     )
                 },
                 output::Format::Json => {
@@ -554,6 +568,8 @@ fn main() -> anyhow::Result<()> {
                                    Ok(())
                                },
                                &mut cache,
+                               // Be quiet.
+                               true, false,
                         );
                     seq.end()?;
                     r
@@ -704,23 +720,43 @@ fn main() -> anyhow::Result<()> {
 
             let mut cache = VerificationCache::new()?;
             let mut vresults = VerificationResult::default();
+
+            let shadow_policy = if let Some(file) = policy_file.path.as_ref() {
+                let policy = std::fs::read(file)
+                    .with_context(|| {
+                        format!("Reading policy file {}", file.display())
+                    })?;
+                Some(policy)
+            } else {
+                None
+            };
+
+            eprintln!("Policy file: {:?}", policy_file.path);
             let result = verify(&git, trust_root,
-                   None,
+                   shadow_policy.as_deref(),
                    (start, new_object),
                    &mut vresults,
                    false,
                    |oid, parent_oid, result| {
                        output::Commit::new(
                            &git, oid, parent_oid, &policy_file.path, result)?
-                           .describe(&mut io::stdout())?;
+                           .describe(&mut io::stdout(), &config.verbosity)?;
                        Ok(())
                    },
                    &mut cache,
+                   config.quiet(), config.verbose(),
             );
 
             let _ = cache.persist();
             result?;
         },
+
+        cli::Subcommand::Version=> {
+            eprintln!("sq-git {}", env!("CARGO_PKG_VERSION"));
+            eprintln!("using sequoia-openpgp {}", sequoia_openpgp::VERSION);
+            eprintln!("with cryptographic backend {}",
+                      sequoia_openpgp::crypto::backend());
+        }
     }
     Ok(())
 }
