@@ -305,28 +305,33 @@ https://github.com/orgs/community/discussions/27607"))
         })
     }
 
-    pub fn describe(&self, sink: &mut dyn io::Write) -> Result<()> {
+    pub fn describe(&self, sink: &mut dyn io::Write, verbosity: &Verbosity) -> Result<()> {
         for r in &self.results {
             let id = if let Some(parent_id) = self.parent_id {
-                format!("{}..{}", parent_id, self.id)
+                format!("Authenticating {} with {}",
+                        self.id, parent_id)
             } else {
                 self.id.to_string()
             };
 
             match r {
                 Err((e, hint)) => {
-                    writeln!(sink, "{}:\n  Error: {}", id, e)?;
-                    if let Some(summary) = self.summary.as_ref() {
-                        writeln!(sink, "  {}", summary)?;
-                    }
-                    if let Some(h) = hint {
-                        writeln!(sink, "\n  Hint: {}", h)?;
+                    if ! verbosity.quiet() {
+                        writeln!(sink, "{}:\n  Error: {}", id, e)?;
+                        if let Some(summary) = self.summary.as_ref() {
+                            writeln!(sink, "  {}", summary)?;
+                        }
+                        if let Some(h) = hint {
+                            writeln!(sink, "\n  Hint: {}", h)?;
+                        }
                     }
                 },
                 Ok(fp) => {
-                    writeln!(sink, "{}:\n  Signer: {}", id, fp)?;
-                    if let Some(summary) = self.summary.as_ref() {
-                        writeln!(sink, "  {}", summary)?;
+                    if verbosity.verbose() {
+                        writeln!(sink, "{}:\n  Signer: {}", id, fp)?;
+                        if let Some(summary) = self.summary.as_ref() {
+                            writeln!(sink, "  {}", summary)?;
+                        }
                     }
                 },
             }
