@@ -1,5 +1,5 @@
 use std::{
-    collections::{BTreeMap, BTreeSet},
+    collections::{BTreeMap, BTreeSet, btree_map::Entry, HashSet},
     env,
     fmt,
     fs,
@@ -19,6 +19,7 @@ use sequoia_openpgp::{
     Cert,
     Fingerprint,
     KeyHandle,
+    Packet,
     cert::{
         amalgamation::ValidAmalgamation,
         prelude::{SubordinateKeyAmalgamation, UserIDAmalgamation},
@@ -33,12 +34,15 @@ use sequoia_openpgp::{
     parse::{stream::*},
     policy::StandardPolicy,
     serialize::Serialize as _,
+    types::SignatureType,
 };
 
 use crate::{
     Error,
     Result,
     utils::prune_cert,
+    utils::serialize_packet,
+    utils::serialize_signature,
 };
 
 /// Whether to trace execution by default (on stderr).
@@ -50,7 +54,7 @@ const TRACE: bool = false;
 /// change is a change from one git commit with a policy embedded into
 /// it to the next commit, which may change the policy, the source, or
 /// both.
-#[derive(Default, Clone, Deserialize, Serialize)]
+#[derive(Default, Clone, Deserialize, Serialize, PartialEq, Eq)]
 pub struct Policy {
     /// Policy version.
     ///
@@ -58,7 +62,7 @@ pub struct Policy {
     /// forward-compatibility, so that we can evolve the policy
     /// language.
     #[serde(default)]
-    pub version: usize,
+    version: usize,
 
     /// Set of commits that is assumed to be good.
     ///
@@ -67,14 +71,14 @@ pub struct Policy {
     ///
     /// To change this set, you need the `audit` right.
     #[serde(default)]
-    pub commit_goodlist: BTreeSet<String>,
+    commit_goodlist: BTreeSet<String>,
 
     /// Set of authorizations.
     ///
     /// The key is a free-form, human-readable identifier for the
     /// authorization.
     #[serde(default)]
-    pub authorization: BTreeMap<String, Authorization>,
+    authorization: BTreeMap<String, Authorization>,
 }
 
 impl Policy {
@@ -99,7 +103,20 @@ impl Policy {
     }
 
     /// Reads the policy from the given path.
-    pub fn read<P: AsRef<Path>>(path: P) -> Result<Policy> {
+    pub fn read_file<P: AsRef<Path>>(path: P) -> Result<Policy> {
+        let path = path.as_ref();
+        let mut f = fs::File::open(path)?;
+
+        let mut s = String::new();
+        f.read_to_string(&mut s)?;
+        let p: Policy =
+            toml::from_str(&s).map_err(|e| Error::StorageError(e.to_string()))?;
+
+        Ok(p)
+    }
+
+    /// Reads the policy from the given path.
+    pub fn read_file_or_default<P: AsRef<Path>>(path: P) -> Result<Policy> {
         let path = path.as_ref();
         let mut f = match fs::File::open(path) {
             Ok(f) => f,
@@ -119,8 +136,11 @@ impl Policy {
     }
 
     /// Reads the policy from the current git working directory.
+    ///
+    /// Default to the empty policy if the work directory doesn't
+    /// contain a policy.
     pub fn read_from_working_dir() -> Result<Policy> {
-        Self::read(&Self::working_dir_policy_file()?)
+        Self::read_file_or_default(&Self::working_dir_policy_file()?)
     }
 
     /// Reads the policy from the given git commit.
@@ -165,6 +185,47 @@ impl Policy {
         self.write(&Self::working_dir_policy_file()?)
     }
 
+    /// The policy version.
+    pub fn version(&self) -> usize {
+        self.version
+    }
+
+    /// Set of commits that is assumed to be good.
+    ///
+    /// The commits will pass verification even if it would fail for
+    /// whatever reason.
+    ///
+    /// To change this set, you need the `audit` right.
+    pub fn commit_goodlist(&self) -> &BTreeSet<String> {
+        &self.commit_goodlist
+    }
+
+    /// Set of commits that is assumed to be good.
+    ///
+    /// The commits will pass verification even if it would fail for
+    /// whatever reason.
+    ///
+    /// To change this set, you need the `audit` right.
+    pub fn commit_goodlist_mut(&mut self) -> &mut BTreeSet<String> {
+        &mut self.commit_goodlist
+    }
+
+    /// Set of authorizations.
+    ///
+    /// The key is a free-form, human-readable identifier for the
+    /// authorization.
+    pub fn authorization(&self) -> &BTreeMap<String, Authorization> {
+        &self.authorization
+    }
+
+    /// Set of authorizations.
+    ///
+    /// The key is a free-form, human-readable identifier for the
+    /// authorization.
+    pub fn authorization_mut(&mut self) -> &mut BTreeMap<String, Authorization> {
+        &mut self.authorization
+    }
+
     /// Computes the difference between this policy and `other`.
     pub fn diff<'f, 't>(&'f self, other: &'t Policy) -> Result<Diff<'f, 't>> {
         let mut changes = Vec::new();
@@ -317,7 +378,8 @@ impl Policy {
 
         let p = &StandardPolicy::new();
         let h = Helper {
-            policy: self,
+            parent_policy: self,
+            child_policy: commit_policy,
             signer_keys,
             primary_uids,
             results: Default::default(),
@@ -384,7 +446,8 @@ impl Policy {
 
 // This fetches keys and computes the validity of the verification.
 struct Helper<'p> {
-    policy: &'p Policy,
+    parent_policy: &'p Policy,
+    child_policy: &'p Policy,
     //signer_userids: &'p mut BTreeSet<openpgp::packet::UserID>,
     signer_keys: &'p mut BTreeSet<openpgp::Fingerprint>,
     primary_uids: &'p mut BTreeSet<UserID>,
@@ -396,9 +459,9 @@ impl Helper<'_> {
         tracer!(TRACE, "VerificationHelper::handle_result");
         match r {
             Ok(sig) => {
-                self.signer_keys.insert(sig.ka.fingerprint());
+                self.signer_keys.insert(sig.ka.key().fingerprint());
 
-                if let Ok(userid) = sig.ka.cert().primary_userid() {
+                if let Ok(userid) = sig.ka.valid_cert().primary_userid() {
                     let u = userid.userid();
                     if ! self.primary_uids.contains(u) {
                         self.primary_uids.insert(u.clone());
@@ -406,8 +469,8 @@ impl Helper<'_> {
                 }
 
                 self.results.push(
-                    Ok((sig.sig.clone(), sig.ka.cert().cert().clone(),
-                        sig.ka.fingerprint().clone())));
+                    Ok((sig.sig.clone(), sig.ka.cert().clone(),
+                        sig.ka.key().fingerprint().clone())));
             },
             Err(e) => {
                 t!("Signature verification failed: {}", e);
@@ -432,6 +495,10 @@ impl Helper<'_> {
                                       error.to_string()),
                     BadSignature { error, .. } =>
                         Error::BadSignature(error.to_string()),
+                    UnknownSignature { sig, .. } =>
+                        Error::BadSignature(sig.error().to_string()),
+                    u =>
+                        Error::BadSignature(u.to_string()),
                 }));
             },
         }
@@ -443,20 +510,154 @@ impl VerificationHelper for Helper<'_> {
         tracer!(TRACE, "VerificationHelper::get_certs");
         t!("get_certs({:?})", ids);
 
-        let mut certs = vec![];
-        for (name, auth) in self.policy.authorization.iter() {
+        let mut matches: BTreeMap<Fingerprint, Cert> = BTreeMap::new();
+
+        let mut parent: BTreeMap<Fingerprint, Vec<RawCert>> = BTreeMap::new();
+
+        for (name, auth) in self.parent_policy.authorization.iter() {
             for cert in auth.certs()? {
                 let cert = cert?;
+
+                // parent contains all of the certificates.
+                let entry = parent.entry(cert.fingerprint());
+                match entry {
+                    Entry::Occupied(mut oe) => {
+                        oe.get_mut().push(cert.clone());
+                    }
+                    Entry::Vacant(ve) => {
+                        ve.insert(vec![ cert.clone() ]);
+                    }
+                }
+
                 if cert.keys().any(
                     |k| ids.iter().any(|i| i.aliases(&k.key_handle())))
                 {
-                    t!("Signature appears to be from {}", name);
-                    certs.push(Cert::try_from(cert)?);
+                    t!("Signature could be from {}", name);
+
+                    let cert = Cert::try_from(cert)?;
+
+                    let entry = matches.entry(cert.fingerprint());
+                    match entry {
+                        Entry::Occupied(mut oe) => {
+                            let c: &mut Cert = oe.get_mut();
+                            *c = c.clone().merge_public(cert)?;
+                        }
+                        Entry::Vacant(ve) => {
+                            ve.insert(cert);
+                        }
+                    }
                 }
             }
         }
 
-        Ok(certs)
+        // We now merge in certificate updates from the child.  That
+        // is, we *don't* add new certificates.
+
+        // Optimize the typical case: the policies are the same and
+        // thus the child has no updates.
+        if self.parent_policy != self.child_policy {
+            // Merge child into parent stripping any revocation
+            // certificates from the child.
+            let merge = |parent: Cert, child: Cert| -> Result<Cert> {
+                let child = child.into_packets().filter(|p| {
+                    if let Packet::Signature(sig) = p {
+                        ! matches!(sig.typ(),
+                                   SignatureType::KeyRevocation
+                                   | SignatureType::SubkeyRevocation
+                                   | SignatureType::CertificationRevocation)
+                    } else {
+                        true
+                    }
+                });
+
+                Ok(parent.insert_packets(child)?.0)
+            };
+
+            for (name, auth) in self.child_policy.authorization.iter() {
+                let child_certs = if let Ok(certs) = auth.certs() {
+                    certs
+                } else {
+                    // Silently ignore any parse errors.
+                    continue;
+                };
+
+                for child_cert in child_certs {
+                    let child_cert = if let Ok(cert) = child_cert {
+                        cert
+                    } else {
+                        // Silently ignore any parse errors.
+                        continue;
+                    };
+
+                    let fpr = child_cert.fingerprint();
+                    if let Some(cert) = matches.get_mut(&fpr) {
+                        t!("Updating {}", fpr);
+
+                        let child_cert = if let Ok(cert)
+                            = Cert::try_from(child_cert)
+                        {
+                            cert
+                        } else {
+                            // Silently ignore any parse errors.
+                            continue;
+                        };
+
+                        if let Ok(merged) = merge(cert.clone(), child_cert) {
+                            *cert = merged;
+                        }
+                    } else {
+                        if child_cert.keys().any(
+                            |k| ids.iter().any(|i| i.aliases(&k.key_handle())))
+                        {
+                            t!("Signature could be from {}", name);
+
+                            // cert is not in matches, but it is
+                            // present in the parent policy.  This
+                            // happens if the certificate gains a
+                            // subkey in the child policy, and the
+                            // commit is signed by that subkey.
+                            if let Some(certs)
+                                = parent.get(&child_cert.fingerprint())
+                            {
+                                if let Ok(child_cert) = Cert::try_from(child_cert) {
+                                    let mut parent: Option<Cert> = None;
+                                    for c in certs.into_iter() {
+                                        let c = if let Ok(c) = Cert::try_from(c) {
+                                            c
+                                        } else {
+                                            continue;
+                                        };
+                                        if let Some(parent) = parent.as_mut() {
+                                            if let Ok(merged) =
+                                                parent.clone().merge_public(c)
+                                            {
+                                                *parent = merged;
+                                            }
+                                        } else {
+                                            parent = Some(c);
+                                        }
+                                    }
+
+                                    let parent = if let Some(parent) = parent {
+                                        parent
+                                    } else {
+                                        continue;
+                                    };
+
+                                    // Merge in the child.
+                                    if let Ok(merged) = merge(parent, child_cert) {
+                                        matches.insert(
+                                            merged.fingerprint(), merged);
+                                    }
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+        }
+
+        Ok(matches.into_values().collect())
     }
     fn check(&mut self, structure: MessageStructure) -> openpgp::Result<()> {
         tracer!(TRACE, "VerificationHelper::get_certs");
@@ -479,7 +680,7 @@ impl VerificationHelper for Helper<'_> {
     }
 }
 
-#[derive(Default, Clone, Deserialize, Serialize)]
+#[derive(Default, Clone, Deserialize, Serialize, PartialEq, Eq)]
 pub struct Authorization {
     #[serde(default, skip_serializing_if = "bool_is_false")]
     pub sign_commit: bool,
@@ -603,7 +804,149 @@ impl Authorization {
             changes.push(Change::AddRight(name.clone(), Right::Audit));
         }
 
-        // XXX compare certs
+        // Check for changes to the certificates.
+        if self.keyring != other.keyring {
+            // Parse the keyring.
+            //
+            // XXX: We silently ignore certificates that we are unable
+            // to parse.  We should perhaps reject any changes that we
+            // don't understand.
+            fn parse_keyring<'a>(name: &str, keyring: &'a str)
+                -> BTreeMap<Fingerprint, Vec<RawCert<'a>>>
+            {
+                match RawCertParser::from_bytes(keyring) {
+                    Err(err) => {
+                        eprintln!("Parsing {}'s keyring: {}", name, err);
+                        Default::default()
+                    }
+                    Ok(certs) => {
+                        let certs = certs.into_iter()
+                            .filter_map(|cert| {
+                                match cert {
+                                    Err(err) => {
+                                        eprintln!("Parsing certificate from {}'s keyring: {}",
+                                                  name, err);
+                                        None
+                                    }
+                                    Ok(cert) => Some(cert),
+                                }
+                            });
+
+                        let mut map: BTreeMap<Fingerprint, Vec<RawCert>> = BTreeMap::new();
+                        for cert in certs {
+                            let entry = map.entry(cert.fingerprint());
+                            match entry {
+                                Entry::Occupied(mut oe) => {
+                                    let oe: &mut Vec<RawCert> = oe.get_mut();
+                                    oe.push(cert);
+                                }
+                                Entry::Vacant(ve) => {
+                                    ve.insert(vec![ cert ]);
+                                }
+                            }
+                        }
+
+                        map
+                    }
+                }
+            }
+
+            let old = parse_keyring(&name, &self.keyring);
+            let new = parse_keyring(&name, &other.keyring);
+
+            // Identify certificates have been added or removed.
+            let old_certs: BTreeSet<&Fingerprint> = old.keys().collect();
+            let new_certs: BTreeSet<&Fingerprint> = new.keys().collect();
+
+            for &removed in old_certs.difference(&new_certs) {
+                changes.push(Change::RemoveCert(name.clone(), removed.clone()));
+            }
+            for &added in new_certs.difference(&old_certs) {
+                changes.push(Change::AddCert(name.clone(), added.clone()));
+            }
+
+            // For certificates that the two keyrings have in common,
+            // identify the packets that have been added or removed.
+            for &fpr in old_certs.intersection(&new_certs) {
+                let old = old.get(fpr).expect("have it");
+                let new = new.get(fpr).expect("have it");
+
+                if old == new {
+                    continue;
+                }
+
+                // Convert each signature into a tuple consisting of a
+                // signature and its associated component.
+                let into_packets = |certs: Vec<RawCert>| -> HashSet<(Packet, Signature)> {
+                    let mut pairs: HashSet<(Packet, Signature)> = HashSet::new();
+
+                    for cert in certs.into_iter() {
+                        let mut packets = cert.packets();
+                        let primary_key
+                            = packets.next().expect("have a primary key");
+                        let primary_key = match Packet::try_from(primary_key) {
+                            Ok(p) => p,
+                            Err(err) => {
+                                eprintln!(
+                                    "Warning: {} has a corrupted primary key \
+                                     packet (skipped): {}",
+                                    cert.fingerprint(), err);
+                                continue;
+                            }
+                        };
+
+                        packets.fold(
+                            primary_key,
+                            |component, packet| {
+                                let packet = match Packet::try_from(packet) {
+                                    Ok(p) => p,
+                                    Err(err) => {
+                                        eprintln!(
+                                            "Warning: {} has a corrupted packet \
+                                             (skipped): {}",
+                                            cert.fingerprint(), err);
+                                        return component;
+                                    }
+                                };
+
+                                match packet {
+                                    Packet::Signature(sig) => {
+                                        pairs.insert(
+                                            (component.clone(), sig.clone()));
+                                        component
+                                    }
+                                    Packet::Marker(_) => {
+                                        // Silently skip marker packets.
+                                        component
+                                    }
+                                    _ => {
+                                        // New component.
+                                        packet.clone()
+                                    }
+                                }
+                            });
+                    }
+
+                    pairs
+                };
+
+                let old_packets = into_packets(old.clone());
+                let new_packets = into_packets(new.clone());
+
+                for (component, sig) in old_packets.difference(&new_packets) {
+                    changes.push(Change::RemovePacket(name.clone(),
+                                                      fpr.clone(),
+                                                      component.clone(),
+                                                      sig.clone()));
+                }
+                for (component, sig) in new_packets.difference(&old_packets) {
+                    changes.push(Change::AddPacket(name.clone(),
+                                                   fpr.clone(),
+                                                   component.clone(),
+                                                   sig.clone()));
+                }
+            }
+        }
     }
 }
 
@@ -651,26 +994,23 @@ pub enum Change {
 
     AddCert(String,
             #[serde(serialize_with = "serialize_fp")] Fingerprint),
-    AddUserID(String,
-              #[serde(serialize_with = "serialize_fp")] Fingerprint,
-              String),
-    AddSubkey(String,
-              #[serde(serialize_with = "serialize_fp")] Fingerprint,
-              #[serde(serialize_with = "serialize_fp")] Fingerprint),
-    AddSignatures(String,
-                  #[serde(serialize_with = "serialize_fp")] Fingerprint,
-                  usize),
     RemoveCert(String,
                #[serde(serialize_with = "serialize_fp")] Fingerprint),
-    RemoveUserID(String,
+
+    AddPacket(String,
+              // The certificate.
+              #[serde(serialize_with = "serialize_fp")] Fingerprint,
+              // The component that the packet is associated with.
+              #[serde(serialize_with = "serialize_packet")] Packet,
+              // The signature that was added.
+              #[serde(serialize_with = "serialize_signature")] Signature),
+    RemovePacket(String,
+                 // The certificate.
                  #[serde(serialize_with = "serialize_fp")] Fingerprint,
-                 String),
-    RemoveSubkey(String,
-                 #[serde(serialize_with = "serialize_fp")] Fingerprint,
-                 #[serde(serialize_with = "serialize_fp")] Fingerprint),
-    RemoveSignatures(String,
-                     #[serde(serialize_with = "serialize_fp")] Fingerprint,
-                     usize),
+                 // The component that the packet is associated with.
+                 #[serde(serialize_with = "serialize_packet")] Packet,
+                 // The signature that was removed.
+                 #[serde(serialize_with = "serialize_signature")] Signature),
 }
 
 impl Change {
@@ -694,14 +1034,20 @@ impl Change {
             RemoveCert(_, _) => r.assert(Right::RetireUser),
 
             // Lenient cert updates.
-            AddUserID(_, _, _) => Ok(()),
-            AddSubkey(_, _, _) => Ok(()),
-            AddSignatures(_, _, _) => Ok(()),
+            AddPacket(_, _, _, _) => Ok(()),
 
             // Strict cert trimmings.
-            RemoveUserID(_, _, _) => r.assert(Right::RetireUser),
-            RemoveSubkey(_, _, _) => r.assert(Right::RetireUser),
-            RemoveSignatures(_, _, _) => r.assert(Right::RetireUser),
+            RemovePacket(_, fpr, _component, sig) => {
+                if sig.get_issuers().into_iter()
+                    .any(|kh| kh.aliases(KeyHandle::from(fpr)))
+                {
+                    // First party signatures can only be removed by
+                    // authorized users.
+                    r.assert(Right::RetireUser)
+                } else {
+                    Ok(())
+                }
+            }
         }
     }
 }
