@@ -1,6 +1,7 @@
 //! Commit-tree traversal and verification.
 
 use std::{
+    borrow::Cow,
     path::{Path, PathBuf},
     collections::{
         BTreeMap,
@@ -33,6 +34,7 @@ use sequoia_openpgp::{
 };
 
 use crate::{
+    Error,
     Policy,
     persistent_set,
 };
@@ -53,8 +55,10 @@ pub fn verify(git: &Repository,
               results: &mut VerificationResult,
               keep_going: bool,
               mut verify_cb: impl FnMut(&Oid, Option<&Oid>, &crate::Result<Vec<crate::Result<(String, Signature, Cert, Fingerprint)>>>) -> crate::Result<()>,
-              cache: &mut VerificationCache)
-              -> Result<()> {
+              cache: &mut VerificationCache,
+              quiet: bool, verbose: bool)
+              -> Result<()>
+{
     tracer!(TRACE, "verify");
     t!("verify(_, {}, {}..{})", trust_root, commit_range.0, commit_range.1);
 
@@ -152,11 +156,20 @@ pub fn verify(git: &Repository,
     // so that it can be processed.
 
     // Return the policy associated with the commit.
-    let read_policy = |commit: &Oid| -> Result<Vec<u8>> {
+    let read_policy = |commit: &Oid| -> Result<Cow<[u8]>> {
         if let Some(p) = &shadow_policy {
-            Ok(p.to_vec())
+            Ok(Cow::Borrowed(p))
         } else {
-            Ok(Policy::read_bytes_from_commit(git, commit)?)
+            match Policy::read_bytes_from_commit(git, commit) {
+                Ok(policy) => Ok(Cow::Owned(policy)),
+                Err(err) => {
+                    if let Error::MissingPolicy(_) = err {
+                        Ok(Cow::Borrowed(b""))
+                    } else {
+                        Err(err.into())
+                    }
+                }
+            }
         }
     };
 
@@ -245,8 +258,9 @@ pub fn verify(git: &Repository,
     };
 
     struct Commit {
-        // Number of children (commits derived from this one).
-        children: usize,
+        // Number of unprocessed children (commits following this
+        // one).
+        unprocessed_children: usize,
 
         // Whether there is an authenticated path from this node to
         // the target.
@@ -259,7 +273,7 @@ pub fn verify(git: &Repository,
     impl Default for Commit {
         fn default() -> Self {
             Commit {
-                children: 0,
+                unprocessed_children: 0,
                 authenticated_suffix: false,
                 traversed_middle: false,
             }
@@ -271,7 +285,7 @@ pub fn verify(git: &Repository,
         commits.insert(
             commit_range.1.clone(),
             Commit {
-                children: 0,
+                unprocessed_children: 0,
                 authenticated_suffix: true,
                 traversed_middle: middle.is_none(),
             });
@@ -279,7 +293,7 @@ pub fn verify(git: &Repository,
 
     // We walk the tree from the target to the trust root (using a
     // breath first search, but it doesn't matter), and fill in
-    // COMMITS.CHILDREN.
+    // COMMITS.
     {
         // Commits that we haven't processed yet.
         let mut pending: BTreeSet<Oid> = Default::default();
@@ -305,8 +319,10 @@ pub fn verify(git: &Repository,
             for parent in commit.parents() {
                 let parent_id = parent.id();
 
+                // Add an entry for the parent commit (if we haven't
+                // yet done so).
                 let info = commits.entry(parent_id).or_default();
-                info.children += 1;
+                info.unprocessed_children += 1;
 
                 if ! processed.contains(&parent_id)
                     && ! pending.contains(&parent_id)
@@ -376,13 +392,15 @@ pub fn verify(git: &Repository,
 
         if cache_hit {
             // XXX: communicate this to the caller
-            // instead of eprintln.
+            // instead of println.
             let id = if let Some(parent_id) = parent_id {
                 format!("{}..{}", parent_id, commit_id)
             } else {
                 commit_id.to_string()
             };
-            eprintln!("{}:\n  Cached positive verification", id);
+            if verbose {
+                println!("{}:\n  Cached positive verification", id);
+            }
         }
 
         match vresult {
@@ -400,7 +418,9 @@ pub fn verify(git: &Repository,
                 if ! cache_hit && results.is_empty() {
                     // XXX: communicate this to the caller
                     // instead of eprintln.
-                    eprintln!("{}: Explicitly goodlisted", commit_id);
+                    if verbose {
+                        println!("{}: Explicitly goodlisted", commit_id);
+                    }
                     good = true;
                 }
 
@@ -502,7 +522,8 @@ pub fn verify(git: &Repository,
     let mut valid_path = trust_root == commit_range.0
         && commit_range.0 == commit_range.1;
 
-    // Commits that we haven't processed yet.
+    // Commits that we haven't processed yet and have no unprocessed
+    // children.
     let mut pending: BTreeSet<Oid> = Default::default();
     if trust_root != commit_range.1 {
         pending.insert(commit_range.1.clone());
@@ -514,19 +535,28 @@ pub fn verify(git: &Repository,
         t!("Processing {}: {}", commit_id, commit.summary().unwrap_or(""));
 
         let commit_info = commits.get(&commit_id).expect("added");
-        assert_eq!(commit_info.children, 0);
+        assert_eq!(commit_info.unprocessed_children, 0);
         let authenticated_suffix = commit_info.authenticated_suffix;
         let traversed_middle = commit_info.traversed_middle;
 
         for (parent_i, parent) in commit.parents().enumerate() {
             let parent_id = parent.id();
-            t!("Considering {} -> {} (parent #{})",
-               commit_id, parent_id, parent_i);
+            t!("Considering {} -> {} (parent #{} of {})",
+               commit_id, parent_id, parent_i + 1, commit.parents().len());
 
-            let parent_info = commits.get_mut(&parent_id).expect("added");
-            t!("  Parent has {} unprocessed children", parent_info.children);
-            assert!(parent_info.children > 0);
-            parent_info.children -= 1;
+            let parent_is_trust_root = parent_id == trust_root;
+
+            let parent_info = commits.get_mut(&parent_id)
+                .with_context(|| format!("Looking up {}", parent_id))
+                .expect("added");
+            t!("  Parent has {} unprocessed children",
+               parent_info.unprocessed_children);
+            assert!(parent_info.unprocessed_children > 0);
+            parent_info.unprocessed_children -= 1;
+            if parent_info.unprocessed_children == 0 && ! parent_is_trust_root {
+                t!("  Adding parent to pending queue for processing");
+                pending.insert(parent_id);
+            }
 
             if authenticated_suffix {
                 t!("  Child IS on an authenticated suffix");
@@ -543,9 +573,14 @@ pub fn verify(git: &Repository,
             } else {
                 false
             };
+            t!("  Could {}authenticate commit.",
+               if authenticated { "" } else { "not " });
 
             if authenticated_suffix && authenticated {
                 t!("  Parent authenticates child");
+                if ! parent_info.authenticated_suffix {
+                    t!("  Parent is now part of an authenticated suffix.");
+                }
                 parent_info.authenticated_suffix = true;
 
                 if traversed_middle {
@@ -555,38 +590,27 @@ pub fn verify(git: &Repository,
                     parent_info.traversed_middle = true;
                 }
 
-                if parent_id == trust_root {
+                if parent_is_trust_root {
                     t!("  Parent is the trust root.");
 
-                    // This is the trust root.  There is no need to go
-                    // further.
                     if ! parent_info.traversed_middle {
                         t!("  but path was not via {}", middle.unwrap());
                     } else {
+                        // This is the trust root and we traversed the
+                        // middle commit.  There is no need to go
+                        // further.
                         valid_path = true;
+                        if ! keep_going {
+                            break 'authentication;
+                        }
                     }
-                    if ! keep_going {
-                        break 'authentication;
-                    }
-                }
-            }
-
-            if parent_info.children == 0 {
-                t!("  No other unprocessed children.");
-                if parent_id == trust_root {
-                    t!("  Parent is the start of the commit range, \
-                        so it doesn't need to be processed.");
-                } else if ! keep_going && ! parent_info.authenticated_suffix {
-                    t!("  Parent does not authenticate any child, \
-                        so it doesn't need to be processed.");
-                } else {
-                    t!("  Adding parent to pending.");
-                    pending.insert(parent_id);
                 }
             }
         }
     }
 
+    t!("No commits pending.");
+
     // When using a shadow policy, we also authenticate the trust
     // root with it.
     if (keep_going || valid_path) && shadow_policy.is_some() {
@@ -610,16 +634,20 @@ pub fn verify(git: &Repository,
 
     if valid_path {
         if trust_root == commit_range.0 {
-            eprintln!(
-                "Verified that there is an authenticated path from the trust root\n\
-                 {} to {}.",
-                trust_root, commit_range.1);
+            if ! quiet {
+                println!(
+                    "Verified that there is an authenticated path from the trust root\n\
+                     {} to {}.",
+                    trust_root, commit_range.1);
+            }
         } else {
-            eprintln!(
-                "Verified that there is an authenticated path from the trust root\n\
-                 {} via {}\n\
-                 to {}.",
-                trust_root, commit_range.0, commit_range.1);
+            if ! quiet {
+                println!(
+                    "Verified that there is an authenticated path from the trust root\n\
+                     {} via {}\n\
+                     to {}.",
+                    trust_root, commit_range.0, commit_range.1);
+            }
         }
         Ok(())
     } else {
