@@ -282,19 +282,7 @@ impl Environment {
         e.import(&e.xander.cert)?;
         e.import(&e.riley.cert)?;
 
-        e.git(&["version"]);
-
-        e.git(&["init", &e.git_state().display().to_string()])?;
-        e.git(&["config", "--local", "user.email", "you@example.org"])?;
-        e.git(&["config", "--local", "user.name", "Your Name"])?;
-
-        // git's default is to not sign.  But, the user might have
-        // overridden this in their ~/.gitconfig, and be using an old
-        // version of git (<2.32).  In that case, GIT_CONFIG_GLOBAL
-        // won't suppress this setting.  Setting it unconditionally in
-        // the local configuration file is a sufficient workaround.
-        e.git(&["config", "--local", "user.signingkey", "0xDEADBEEF"])?;
-        e.git(&["config", "--local", "commit.gpgsign", "false"])?;
+        e.init_repo(e.git_state().as_path())?;
 
         Ok(e)
     }
@@ -311,43 +299,129 @@ impl Environment {
                   self.wd.path().display());
     }
 
+    pub fn init_repo<'a, P: Into<Option<&'a Path>>>(&self, wd: P) -> Result<()> {
+        let git_state = self.git_state();
+        let wd = wd.into().unwrap_or(git_state.as_path());
+
+        self.git(&["version"])?;
+
+        eprintln!("Creating {}", wd.display());
+        std::fs::create_dir(wd)
+            .or_else(|err| {
+                if err.kind() == std::io::ErrorKind::AlreadyExists {
+                    // If it already exists, don't fail.
+                    Ok(())
+                } else {
+                    Err(err)
+                }
+            })
+            .expect(&format!("Can create {}", wd.display()));
+
+        self.git(&["init", &wd.display().to_string()])?;
+        self.git_in(wd, &["config", "--local", "user.email", "you@example.org"])?;
+        self.git_in(wd, &["config", "--local", "user.name", "Your Name"])?;
+
+        // git's default is to not sign.  But, the user might have
+        // overridden this in their ~/.gitconfig, and be using an old
+        // version of git (<2.32).  In that case, GIT_CONFIG_GLOBAL
+        // won't suppress this setting.  Setting it unconditionally in
+        // the local configuration file is a sufficient workaround.
+        self.git_in(wd, &["config", "--local", "user.signingkey", "0xDEADBEEF"])?;
+        self.git_in(wd, &["config", "--local", "commit.gpgsign", "false"])?;
+
+        // Make sure sq-git works.
+        self.sq_git_in(wd, &["version"])?;
+
+        Ok(())
+    }
+
     /// Returns the environment and the root commit.
-    #[allow(dead_code)]
-    pub fn scooby_gang_bootstrap() -> Result<(Environment, String)> {
+    ///
+    /// The initial commit is signed by the specified identity
+    /// (default: "Willow Rosenberg Release Signing").
+    pub fn scooby_gang_bootstrap(signer: Option<&str>) -> Result<(Environment, String)> {
         let e = Environment::new()?;
 
+        e.scooby_gang_bootstrap_in(None, signer)
+            .map(|root| (e, root))
+    }
+
+    #[allow(dead_code)]
+    pub fn scooby_gang_bootstrap_in<'a, P: Into<Option<&'a Path>>>(
+        &self, wd: P, signer: Option<&str>)
+        -> Result<String>
+    {
+        let git_state = self.git_state();
+        let wd = wd.into().unwrap_or(git_state.as_path());
+
+        self.scooby_gang_bootstrap_policy(wd.join("openpgp-policy.toml").as_path());
+
+        let signer = if let Some(signer) = signer {
+            if self.willow.petname == signer {
+                &self.willow.fingerprint
+            } else if self.willow_release.petname == signer {
+                &self.willow_release.fingerprint
+            } else if self.buffy.petname == signer {
+                &self.buffy.fingerprint
+            } else if self.xander.petname == signer {
+                &self.xander.fingerprint
+            } else if self.riley.petname == signer {
+                &self.riley.fingerprint
+            } else {
+                panic!("Unknown signer: {}", signer);
+            }
+        } else {
+            &self.willow_release.fingerprint
+        };
+
+        self.git_in(wd, &["add", "openpgp-policy.toml"])?;
+        self.git_in(
+            wd,
+            &[
+                "commit",
+                "-m", "Initial commit.",
+                &format!("-S{}", signer),
+            ])?;
+        let root = self.git_current_commit_in(wd)?;
+        Ok(root)
+    }
+
+    // Write the policy to the specified file.
+    pub fn scooby_gang_bootstrap_policy(
+        &self, external_policy_file: &Path)
+        -> Result<()>
+    {
+        let external_policy_file = external_policy_file.display().to_string();
+
         // Willow has a code-signing key.
-        e.sq_git(&[
-            "policy",
-            "authorize",
-            e.willow.petname,
-            &e.willow.fingerprint.to_string(),
-            "--sign-commit"
-        ])?;
+        self.sq_git(
+            &[
+                "policy",
+                "authorize",
+                "--policy-file", &external_policy_file,
+                self.willow.petname,
+                &self.willow.fingerprint.to_string(),
+                "--sign-commit"
+            ])?;
 
         // Additionally, Willow also has a release key on her security
         // token.
-        e.sq_git(&[
-            "policy",
-            "authorize",
-            e.willow_release.petname,
-            &e.willow_release.fingerprint.to_string(),
-            "--sign-commit",
-            "--sign-tag",
-            "--sign-archive",
-            "--add-user",
-            "--retire-user",
-            "--audit",
-        ])?;
+        self.sq_git(
+            &[
+                "policy",
+                "authorize",
+                "--policy-file", &external_policy_file,
+                self.willow_release.petname,
+                &self.willow_release.fingerprint.to_string(),
+                "--sign-commit",
+                "--sign-tag",
+                "--sign-archive",
+                "--add-user",
+                "--retire-user",
+                "--audit",
+            ])?;
 
-        e.git(&["add", "openpgp-policy.toml"])?;
-        e.git(&[
-            "commit",
-            "-m", "Initial commit.",
-            &format!("-S{}", e.willow_release.fingerprint),
-        ])?;
-        let root = e.git_current_commit()?;
-        Ok((e, root))
+        Ok(())
     }
 
     pub fn gnupg_state(&self) -> PathBuf {
@@ -468,11 +542,18 @@ impl Environment {
     }
 
     pub fn git<A: AsRef<str>>(&self, args: &[A]) -> Result<Output> {
+        self.git_in(None, args)
+    }
+
+    pub fn git_in<'a, P: Into<Option<&'a Path>>, A: AsRef<str>>(
+        &self, wd: P, args: &[A])
+        -> Result<Output>
+    {
         let mut c = self.command("git");
         for a in args {
             c.arg(a.as_ref());
         }
-        self.run(c)
+        self.run_in(wd, c)
     }
 
     // A convenience function to optionally modify and commit a few
@@ -486,13 +567,25 @@ impl Environment {
                       signer: Option<&Identity>)
         -> Result<String>
     {
-        let p = self.git_state();
+        self.git_commit_in(None, files, commit_msg, signer)
+    }
+
+    pub fn git_commit_in<'a, P: Into<Option<&'a Path>>>(
+        &self,
+        wd: P,
+        files: &[(&str, Option<&[u8]>)],
+        commit_msg: &str,
+        signer: Option<&Identity>)
+        -> Result<String>
+    {
+        let git_state = self.git_state();
+        let wd = wd.into().unwrap_or(git_state.as_path());
 
         for (filename, content) in files.iter() {
             if let Some(content) = content {
-                std::fs::write(p.join(filename), content).unwrap();
+                std::fs::write(wd.join(filename), content).unwrap();
             }
-            self.git(&["add", filename])?;
+            self.git_in(wd, &["add", filename])?;
         }
 
         let mut git_args = vec!["commit", "-m", commit_msg];
@@ -501,40 +594,71 @@ impl Environment {
             signer_ = format!("-S{}", signer.fingerprint);
             git_args.push(&signer_);
         }
-        self.git(&git_args)?;
+        self.git_in(wd, &git_args)?;
+
+        Ok(self.git_current_commit_in(wd)?)
+    }
+
+    // A convenience function to optionally modify and commit a few
+    // files.
+    //
+    // Runs git merge --no-ff
+    #[allow(dead_code)]
+    pub fn git_merge_no_ff(&self,
+                           commits: &[&str],
+                           commit_msg: &str,
+                           signer: Option<&Identity>,
+                           args: &[&str])
+        -> Result<String>
+    {
+        let p = self.git_state();
+
+        let mut git_args = vec!["merge", "--no-ff", "-m", commit_msg];
+        git_args.extend_from_slice(args);
+        let signer_;
+        if let Some(signer) = signer {
+            signer_ = format!("-S{}", signer.fingerprint);
+            git_args.push(&signer_);
+        }
+
+        git_args.extend_from_slice(commits);
+
+        let output = self.git(&git_args)?;
+        assert!(!output.stdout.starts_with(&b"Already up to date."[..]),
+                "Failed to create merge commit.  Perhaps you are merging \
+                 commits that are already merged.");
 
         Ok(self.git_current_commit()?)
     }
 
     pub fn git_current_commit(&self) -> Result<String> {
-        Ok(String::from_utf8(self.git(&["rev-parse", "HEAD"])?.stdout)?
+        self.git_current_commit_in(None)
+    }
+
+    pub fn git_current_commit_in<'a, P: Into<Option<&'a Path>>>(&self, wd: P)
+        -> Result<String>
+        {
+        Ok(String::from_utf8(self.git_in(wd, &["rev-parse", "HEAD"])?.stdout)?
+           .trim().to_string())
+    }
+
+    pub fn git_log_graph(&self) -> Result<String> {
+        Ok(String::from_utf8(self.git(&["log", "--graph", "--show-signature"])?.stdout)?
            .trim().to_string())
     }
 
     pub fn sq_git_path() -> Result<PathBuf> {
-        use std::sync::Once;
-        static BUILD: Once = Once::new();
-        BUILD.call_once(|| {
-            let o = Command::new("cargo")
-                .arg("build").arg("--quiet")
-                .arg("--bin").arg("sq-git")
-                .output()
-                .expect("running cargo failed");
-            if ! o.status.success() {
-                panic!("build failed:\n\nstdout:\n{}\n\n stderr:\n{}",
-                        String::from_utf8_lossy(&o.stdout),
-                        String::from_utf8_lossy(&o.stderr));
-            }
-        });
-
-        Ok(if let Ok(target) = std::env::var("CARGO_TARGET_DIR") {
-            PathBuf::from(target).canonicalize()?
-        } else {
-            std::env::current_dir()?.join("target")
-        }.join("debug/sq-git"))
+        Ok(env!("CARGO_BIN_EXE_sq-git").into())
     }
 
     pub fn sq_git<A: AsRef<str>>(&self, args: &[A]) -> std::result::Result<Output, Error> {
+        self.sq_git_in(None, args)
+    }
+
+    pub fn sq_git_in<'a, P: Into<Option<&'a Path>>, A: AsRef<str>>(
+        &self, wd: P, args: &[A])
+        -> std::result::Result<Output, Error>
+    {
         let mut c = self.command(Self::sq_git_path()?);
 
         // We are a machine, request machine-readable output.
@@ -544,10 +668,19 @@ impl Environment {
             c.arg(a.as_ref());
         }
 
-        self.run(c)
+        self.run_in(wd, c)
     }
 
     pub fn spawn(&self, mut c: Command) -> Result<Child> {
+        self.spawn_in(None, c)
+    }
+
+    pub fn spawn_in<'a, P: Into<Option<&'a Path>>>(&self, wd: P, mut c: Command)
+        -> Result<Child>
+    {
+        let git_state = self.git_state();
+        let wd = wd.into().unwrap_or(git_state.as_path());
+
         // Preserve any environment variables the caller explicitly
         // set (env_clear clears those and prevents the process from
         // inheriting any environment variables, but we only want the
@@ -562,7 +695,7 @@ impl Environment {
             })
             .collect::<Vec<(OsString, OsString)>>();
 
-        let c = c.current_dir(self.git_state())
+        let c = c.current_dir(wd)
             .env_clear() // Filter out all git-related environment variables.
             .envs(std::env::vars()
                   .filter(|(k, _)| ! k.starts_with("GIT_"))
@@ -581,11 +714,26 @@ impl Environment {
     }
 
     pub fn run(&self, c: Command) -> Result<Output> {
+        self.run_in(None, c)
+    }
+
+    pub fn run_in<'a, P: Into<Option<&'a Path>>>(&self, wd: P, c: Command)
+        -> Result<Output>
+    {
         let cmd = format!("{:?}", c);
 
         eprintln!("$ {}", cmd);
 
-        let output = self.spawn(c)?.wait_with_output()?;
+        let output = self.spawn_in(wd, c)
+            .map_err(|err| {
+                eprintln!(" -> failed to spawn command: {}", err);
+                err
+            })?
+            .wait_with_output()
+            .map_err(|err| {
+                eprintln!(" -> waiting for command to complete: {}", err);
+                err
+            })?;
 
         if output.status.success() {
             eprintln!(" -> success");
