@@ -4,7 +4,7 @@ use common::Environment;
 use common::Result;
 
 fn create_environment() -> Result<(Environment, String)> {
-    Environment::scooby_gang_bootstrap()
+    Environment::scooby_gang_bootstrap(None)
 }
 
 // The keys for the different authorizations in the JSON file.
