@@ -3,7 +3,7 @@ use common::Environment;
 use common::Result;
 
 fn create_environment() -> Result<(Environment, String)> {
-    Environment::scooby_gang_bootstrap()
+    Environment::scooby_gang_bootstrap(None)
 }
 
 #[test]
